module trigger(
    input                clk,
    input                reset,
    
    input                adc_val,
    output               adc_rdy,
    input       [47:0]   adc_data,

    output wire [63:0]   buf_writedata,
    output reg  [12:0]   buf_address,
    output wire          buf_write,
    input                buf_ack,

    input                refresh_rdy,
    output               refresh_val,
	 
	 input  wire			 trig_inhibit,

    input       [31:0]  cfg
);

localparam ST_ARMED        = 3'd0;
localparam ST_READY        = 3'd1;
localparam ST_TRIGGERED    = 3'd2;
localparam ST_FLUSH_LEFT   = 3'd3;
localparam ST_FLUSH_RIGHT  = 3'd4;
localparam ST_SEND_REFRESH = 3'd5;

reg [2:0]  state, state_next;
reg [8:0]  right_index, right_index_next;
reg [8:0]  left_index, left_index_next;
reg [8:0]  left_origin, left_origin_next;
reg [12:0] buf_address_next;

wire left_we, right_we;
wire [47:0] left_data, right_data;
wire [47:0] flush_data;
wire [11:0] math_data;

wire math_mode;
wire [1:0] math_in0, math_in1;
wire [11:0] math_in0_data, math_in1_data;
wire [11:0] channel_data [3:0];

assign channel_data[0] = flush_data[11:0];
assign channel_data[1] = flush_data[23:12];
assign channel_data[2] = flush_data[35:24];
assign channel_data[3] = flush_data[47:36];

cfg_reg #(1,20,1'b0) math_mode_reg(clk,reset,cfg,math_mode);
cfg_reg #(2,21,2'b0) math_in0_reg(clk,reset,cfg,math_in0);
cfg_reg #(2,22,2'b0) math_in1_reg(clk,reset,cfg,math_in1);

assign math_in0_data = {1'b0, channel_data[math_in0][11:1]};
assign math_in1_data = {1'b0, channel_data[math_in1][11:1]};

assign math_data = math_mode ? (12'h800 + math_in0_data - math_in1_data) : math_in0_data + math_in1_data;

wire triggered;

triggerer trig(
    .clk(clk),
    .reset(reset),
    
    .data_val(left_we),
    .new_data(adc_data),

    .cfg(cfg),

    .triggered(triggered)
);



single_port_ram left_buffer(
    .data(adc_data),
    .addr(left_index),
    .we(left_we), 
    .clk(clk),
    .q(left_data)
);

single_port_ram right_buffer(
	 .data(adc_data),
    .addr(right_index),
    .we(right_we), 
    .clk(clk),
    .q(right_data)
);

assign adc_rdy = (state == ST_ARMED) | (state == ST_READY) | (state == ST_TRIGGERED);

assign left_we = adc_val & ((state == ST_ARMED) | (state == ST_READY));
assign right_we = adc_val & (state == ST_TRIGGERED);

assign buf_write = (state == ST_FLUSH_LEFT) | (state == ST_FLUSH_RIGHT);
assign flush_data = (state == ST_FLUSH_LEFT) ? left_data : right_data;
assign buf_writedata = {math_data,flush_data};


assign refresh_val = (state == ST_SEND_REFRESH);


// Next state
always @(*) begin
    state_next = state;
    left_index_next = left_index;
    left_origin_next = left_origin;
    right_index_next = right_index;
    buf_address_next = buf_address;
    
    case(state)
    ST_ARMED: begin
        state_next = (left_index == 9'd319) ? ST_READY : ST_ARMED;
        left_origin_next = 9'd0;
        if(adc_val) begin
            left_index_next = left_index + 9'd1;
        end
    end
    ST_READY: begin
        state_next  = (triggered & ~trig_inhibit) ? ST_TRIGGERED : ST_READY;
        if(adc_val) begin
            if (left_index == 9'd319)
                left_index_next = 9'd0;
            else
                left_index_next = left_index  + 9'd1;
                
            if (left_origin == 9'd319)
                left_origin_next = 9'd0;
            else
                left_origin_next = left_origin + 9'd1;
        end
    end
    ST_TRIGGERED: begin
        state_next = (right_index == 9'd319) ? ST_FLUSH_LEFT : ST_TRIGGERED;
        buf_address_next = 13'd0;
        left_index_next = left_origin;
        if(adc_val) begin
            right_index_next = right_index + 9'd1;
        end
    end
    ST_FLUSH_LEFT: begin
        right_index_next = 9'd0;
        if(buf_ack) begin
             state_next = (buf_address == 13'd2552) ? ST_FLUSH_RIGHT : ST_FLUSH_LEFT;
		  if (left_index == 9'd319)
                left_index_next = 9'd0;
            else
                left_index_next = left_index  + 9'd1;
                
            buf_address_next = buf_address + 13'd8;
        end
    end
    ST_FLUSH_RIGHT: begin
        if(buf_ack) begin
				state_next = (buf_address == 13'd5112) ? ST_SEND_REFRESH : ST_FLUSH_RIGHT;
				right_index_next = right_index + 9'd1;
            buf_address_next = buf_address + 13'd8;
        end
    end
    ST_SEND_REFRESH: begin
        state_next = refresh_rdy ? ST_ARMED : ST_SEND_REFRESH;
        left_index_next  = 9'd0;
        left_origin_next = 9'd0;
        right_index_next = 9'd0;
    end
    endcase
end

always @(posedge clk) begin
    if (reset) begin
        state       <= ST_ARMED;
        left_index  <= 9'd0;
       left_origin <= 9'd0;
       right_index <= 9'd0;
       buf_address <= 13'd0;
    end else begin
        state       <= state_next;
        left_index  <= left_index_next;
       left_origin <= left_origin_next;
       right_index <= right_index_next;
       buf_address <= buf_address_next;
    end
end


endmodule
