module triggerer(
    input           clk,
    input           reset,


    input           data_val,
    input   [47:0]  new_data, 

    input   [31:0]  cfg,

    output  reg     triggered
);


wire     [3:0]  trig_mode;
wire            trig_en;
wire    	[11:0] trig_level;
wire    	[9:0]  trig_pulse_width;
wire 		[1:0]  trig_sel;

localparam  MODE_RISE      = 4'd0;
localparam  MODE_FALL      = 4'd1;
localparam  MODE_PULSE_EQ  = 4'd2;
localparam  MODE_PULSE_LT  = 4'd3;
localparam  MODE_PULSE_GT  = 4'd4;

cfg_reg #(4,5,0)  cfg_5(clk,reset,cfg,trig_mode); 
cfg_reg #(1,6,0)  cfg_6(clk,reset,cfg,trig_en);
cfg_reg #(12,7,12'h800) cfg_7(clk,reset,cfg,trig_level);
cfg_reg #(10,8,0) cfg_8(clk,reset,cfg,trig_pulse_width);
cfg_reg #(2,9,0)  cfg_9(clk,reset,cfg,trig_sel);


reg above_level;

always@(*)begin
	case(trig_sel)
		2'b00: above_level = new_data[11:0]  > trig_level;
		2'b01: above_level = new_data[23:12] > trig_level;
		2'b10: above_level = new_data[35:24] > trig_level;
		2'b11: above_level = new_data[47:36] > trig_level;
	endcase
end


//-----------------------------------------------------------------
// Trigger Selection
//-----------------------------------------------------------------

wire rise_trig;
wire fall_trig;
wire pulse_eq_trig;
wire pulse_lt_trig;
wire pulse_gt_trig;

always @(*) begin
    if(trig_en) begin
        case(trig_mode)
            MODE_RISE: triggered = rise_trig;
            MODE_FALL: triggered = fall_trig;
            MODE_PULSE_EQ: triggered = pulse_eq_trig;
            MODE_PULSE_LT: triggered = pulse_lt_trig;
            MODE_PULSE_GT: triggered = pulse_gt_trig;
            default:       triggered = 1'b0;
        endcase
    end else begin
        triggered = 1'b1;
    end
end

//------------------------------------------------------------------
// Edge Triggering
//------------------------------------------------------------------

reg past_above_level, past_above_level_nxt;

assign rise_trig = above_level & !past_above_level; 
assign fall_trig = !above_level & past_above_level;


always @(*) begin
    past_above_level_nxt = data_val ? above_level : past_above_level;
end

always @(posedge clk) begin
    if(reset) begin
        past_above_level <= 0;
    end else begin
        past_above_level <= past_above_level_nxt;
    end
end



//------------------------------------------------------------------
// Pulse Triggering
//------------------------------------------------------------------


reg         pulse_active, pulse_active_nxt;
reg [9:0]   pulse_size, pulse_size_nxt;
reg         pulse_done;


assign pulse_eq_trig = pulse_done & (pulse_size == trig_pulse_width);
assign pulse_lt_trig = pulse_done & (pulse_size < trig_pulse_width);
assign pulse_gt_trig = pulse_done & (pulse_size > trig_pulse_width);

always @(*) begin
    pulse_active_nxt = pulse_active;
    pulse_size_nxt = pulse_size;
	 pulse_done = 1'b0;
    if(data_val) begin
        if(above_level) begin
            pulse_active_nxt = 1'b1;
            pulse_size_nxt = pulse_size + 10'b1;
        end else begin
            pulse_active_nxt = 1'b0;
            pulse_size_nxt = 10'd0;
            pulse_done = pulse_active;
        end        
    end
end

always @(posedge clk) begin
    if(reset) begin
        pulse_active <= 0;
        pulse_size <= 10'd0;
    end else begin
        pulse_active <= pulse_active_nxt;
        pulse_size   <= pulse_size_nxt;
    end
end

endmodule
