module vga_writer #(
    parameter SCREEN_WIDTH = 640,
    parameter SCREEN_HEIGHT = 480
)(
    input wire clk,
    input wire reset,

    input wire go_val,
    output reg go_rdy,

    output wire [12:0] point_address,
    output reg         point_read,
    input  wire [63:0] point_readdata,
    input  wire        point_acknowledge,

    output wire [18:0] vga_address,
    output reg         vga_write,
    output reg  [7:0]  vga_writedata,

    output wire	       idle,

    input  wire [31:0] cfg
);

localparam ST_WAIT    = 2'd0;
localparam ST_FETCH   = 2'd1;
localparam ST_PROCESS = 2'd2;
localparam ST_DRAW    = 2'd3;
localparam POINT_BASE = 12'h0000_0000;

wire [1:0]	num_channels;

cfg_reg #(2,4,2'd0) num_channels_reg(clk,reset,cfg,num_channels);

reg [63:0] point_values, point_values_next;

wire signed [10:0] incoming_point_y_coord [4:0];
reg  signed [10:0] point_y_coord [4:0];
reg  signed [10:0] point_y_coord_next [4:0];
reg  signed [10:0] prev_y_coord [4:0];
reg  signed [10:0] prev_y_coord_next [4:0];
wire        [4:0]  should_draw_line;

genvar i;
generate
for (i = 0; i < 5; i = i + 1) begin: PLOT_CHANNELS
    wire [17:0] point_scaler;
    cfg_reg #(18, 10+(2*i), 18'd4) point_scaler_reg (clk, reset, cfg, point_scaler);
    wire [10:0] point_offset;
    cfg_reg #(11, 11+(2*i), 11'd0) point_offset_reg (clk, reset, cfg, point_offset);

    wire [11:0] point_value;
    assign point_value = point_values[12*i +: 12];

    wire [17:0] point_mult;
    wire [15:0] point_scaled;
    assign point_mult = {6'b0,point_value} * point_scaler;
    assign point_scaled = point_mult[17:2];

    wire [12:0] point_scaled_saturated;
    assign point_scaled_saturated = (point_scaled > 15'h1FFF) ? 13'h1FFF : point_scaled[12:0];

    assign incoming_point_y_coord[i] = {2'b0, SCREEN_HEIGHT[8:5], 5'b0} 
    - {1'b0, point_scaled_saturated[12:3]} 
    - point_offset;


    assign should_draw_line[i] = ((x_coord != 10'd0) && 
        ((y_coord > prev_y_coord[i]) && (y_coord <= point_y_coord[i])) ||
        ((y_coord < prev_y_coord[i]) && (y_coord >= point_y_coord[i])))
        || (y_coord == point_y_coord[i]);
    end
    endgenerate

    reg [1:0] state, state_next;
    assign idle = state == ST_WAIT;
    reg [9:0] x_coord, x_coord_next;
    reg signed [10:0] y_coord, y_coord_next;

    assign point_address = POINT_BASE + {x_coord, 3'b0};

    assign vga_address = (y_coord[8:0])*SCREEN_WIDTH + x_coord;


    // Registers to configure the line color palette
    wire [7:0] line_color [4:0];
    cfg_reg #(8,30,8'b001_110_11) line_color_reg_0(clk,reset,cfg,line_color[0]);
    cfg_reg #(8,31,8'b111_011_00) line_color_reg_1(clk,reset,cfg,line_color[1]);
    cfg_reg #(8,32,8'b010_111_00) line_color_reg_2(clk,reset,cfg,line_color[2]);
    cfg_reg #(8,33,8'b110_001_11) line_color_reg_3(clk,reset,cfg,line_color[3]);
    cfg_reg #(8,34,8'b111_000_00) line_color_reg_math(clk,reset,cfg,line_color[4]);

    wire math_en;
    cfg_reg #(1,3,1'b0) math_en_reg(clk,reset,cfg,math_en);

    integer j;

    // VGA color selection
    always @(*) begin
        if (should_draw_line)
            vga_writedata = 8'b001_110_11;
        if ((x_coord[4:0] == 5'd0 && x_coord != 10'd0) ||
            (y_coord[4:0] == 5'd0 && y_coord != 9'd0))
            vga_writedata = 8'b010_010_01;
            else
                vga_writedata = 8'd0;

        for (j = 0; j < 4; j = j + 1) begin
            if (j[1:0] <= num_channels && should_draw_line[j]) begin
                vga_writedata = line_color[j];
            end
        end
        if (math_en && should_draw_line[4]) begin
            vga_writedata = line_color[4];
        end
    end

    // Update outputs
    always @(*) begin
        go_rdy     = 1'b0;
        point_read = 1'b0;
        vga_write  = 1'b0;

        case (state)
            ST_WAIT: begin
                go_rdy = 1'b1;
            end
            ST_FETCH: begin
                point_read = 1'b1;
            end
            ST_DRAW: begin
                vga_write = 1'b1;
            end
            default: begin
            end
        endcase
    end


    // Set next state
    always @(*) begin
        state_next = state;
        x_coord_next = x_coord;
        y_coord_next = y_coord;
        point_values_next = point_values;
        for (j = 0; j < 5; j = j + 1) begin
            point_y_coord_next[j] = point_y_coord[j];
            prev_y_coord_next[j] = prev_y_coord[j];
        end

        case (state)
            ST_WAIT: begin
                if (go_val) begin
                    state_next = ST_FETCH;
                    x_coord_next = 10'd0;
                end
                for (j = 0; j < 5; j = j + 1) begin
                    prev_y_coord_next[j] = 9'd0;
                end
            end
            ST_FETCH: begin
                point_values_next = point_readdata;
                if (point_acknowledge) state_next = ST_PROCESS;
            end
        ST_PROCESS: begin
            for (j = 0; j < 5; j = j + 1) begin
                point_y_coord_next[j] = incoming_point_y_coord[j];
            end
            y_coord_next = 9'd0;
            state_next = ST_DRAW;
        end
        ST_DRAW: begin
            if (y_coord == SCREEN_HEIGHT-1) begin
                if (x_coord == SCREEN_WIDTH-1) begin
                    state_next = ST_WAIT;
                end else begin
                    x_coord_next = x_coord + 10'd1;
                    state_next = ST_FETCH;
                    for (j = 0; j < 5; j = j + 1) begin
                        prev_y_coord_next[j] = point_y_coord[j];
                    end
                end
            end else begin
                y_coord_next = y_coord + 9'd1;
            end
        end
        default: begin
            state_next = ST_WAIT;
            x_coord_next = 10'd0;
            y_coord_next = 9'd0;
            for (j = 0; j < 5; j = j + 1) begin
                point_y_coord_next[j] = 9'd0;
            end
        end
    endcase
end

// Flops
always @(posedge clk) begin
    if (reset) begin
        state <= ST_WAIT;
        x_coord <= 10'd0;
        y_coord <= 9'd0;
        point_values <= 64'd0;
        for (j = 0; j < 5; j = i + 1) begin
            point_y_coord[j] <= 9'd0;
            prev_y_coord[j] <= 9'd0;
        end
    end else begin
        state <= state_next;
        x_coord <= x_coord_next;
        y_coord <= y_coord_next;
        point_values <= point_values_next;
        for (j = 0; j < 5; j = j + 1) begin
            point_y_coord[j] <= point_y_coord_next[j];
            prev_y_coord[j] <= prev_y_coord_next[j];
        end
    end
end

endmodule
